'''
Provides classes for loading chunk files from local storage and
putting them out into local storage.

This software is released under an MIT/X11 open source license.

Copyright 2012-2013 Diffeo, Inc.
'''
import os
import sys
import time
import errno
import shutil
import hashlib
import logging
import traceback
import streamcorpus
from cStringIO import StringIO
from _get_name_info import get_name_info
from _tarball_export import tarball_export

logger = logging.getLogger(__name__)

_message_versions = {
    'v0_1_0': streamcorpus.StreamItem_v0_1_0,
    'v0_2_0': streamcorpus.StreamItem_v0_2_0,
    'v0_3_0': streamcorpus.StreamItem_v0_3_0,
    }


class from_local_chunks(object):
    '''
    may use config['max_retries'] (default 1)
    may use config['max_backoff'] (seconds, default 300)
    may use config['streamcorpus_version'] (default 'v0_3_0')
    '''
    def __init__(self, config):
        self.config = config

    def __call__(self, i_str):
        backoff = 0.1
        start_time = time.time()
        tries = 0
        max_retries = int(self.config.get('max_retries', 1))
        last_exc = None
        while tries < max_retries:
            try:
                message = _message_versions[ self.config.get('streamcorpus_version', 'v0_3_0') ]
                logger.debug('reading from %r' % i_str)
                chunk = streamcorpus.Chunk(path=i_str, mode='rb', message=message)
                return chunk
            except IOError, exc:
                if exc.errno == errno.ENOENT:
                    logger.critical('File is missing?  Assume is slow NFS, try %d more times',
                                    max_retries - tries)
                    backoff *= 2
                    tries += 1
                    elapsed = time.time() - start_time
                    if elapsed > self.config.get('max_backoff', 300):
                        ## give up after five minutes of retries
                        logger.critical('File %r not found after %d retries', i_str, tries)
                        raise
                    time.sleep(backoff)
                    last_exc = exc
                else:
                    logger.critical('failed loading %r', i_str, exc_info=True)
                    raise

        ## exceeded max_retries
        logger.critical('File %r not found after %d retries', i_str, tries)
        if last_exc:
            raise last_exc
        raise IOError('File %r not found after %d retries' % (i_str, tries))


def patient_move(path1, path2, max_tries=30):
    backoff = 0.1
    tries = 0
    while tries < max_tries:
        try:
            shutil.copy2(path1, path2)
        except IOError, exc:
            if exc.errno == 2:
                tries += 1
                if tries > max_tries:
                    logger.critical('%d tries failed copying %r to %r', tries, path1, path2, exc_info=True)
                    raise
                logger.critical('attempting retry on shutil.copy2(%r, %r)', path1, path2)
                backoff *= 2
                time.sleep(backoff)
                continue
            else:
                logger.critical('failed copy %r -> %r', path1, path2, exc_info=True)
                raise

        try:
            os.remove(path1)
        except Exception, exc:
            logger.critical('ignoring failure to os.remove(%r)', path1)
            pass

        ## if we get here, we succeeded
        return

    # should never get here
    raise Exception('weird error inside patient_move(%r, %r)' % (path1, path2))


class to_local_chunks(object):
    '''
    uses config dict parts:
    output_name: name of file to write to. may be 'input' to derive name from source path
    output_type:
      'samedir' place files next to input
      'inplace' replace input file with output
      'otherdir' place output files according to config['output_path']
    output_path: where to put output files
    compress: bool() True append '.xz' to output file name and compress output
    '''

    def __init__(self, config):
        self.config = config

    def __call__(self, t_path, name_info, i_str):
        o_type = self.config['output_type']
        
        name_info.update( get_name_info( t_path, i_str=i_str ) )

        if name_info['num'] == 0:
            return None

        if 'input' in self.config['output_name']:
            i_fname = i_str.split('/')[-1]
            if i_fname.endswith('.gpg'):
                i_fname = i_fname[:-4]
            if i_fname.endswith('.xz'):
                i_fname = i_fname[:-3]
            if i_fname.endswith('.sc'):
                i_fname = i_fname[:-3]
            name_info['input_fname'] = i_fname 

        ## prepare to compress the output
        compress = self.config.get('compress', None)
        assert compress in [None, 'xz'], compress

        if o_type == 'samedir':
            ## assume that i_str was a local path
            assert i_str[-3:] == '.sc', repr(i_str[-3:])
            o_path = i_str[:-3] + '-%s.sc' % self.config['output_name']
            if compress:
                o_path += '.xz'
            #print 'creating %s' % o_path
            
        elif o_type == 'inplace':
            ## replace the input chunks with the newly created
            o_path = i_str
            if o_path.endswith('.xz'):
                compress = True

        elif o_type == 'otherdir':
            ## put the 
            if not self.config['output_path'].startswith('/'):
                o_dir = os.path.join(os.getcwd(), self.config['output_path'])
            else:
                o_dir = self.config['output_path']

            if not os.path.exists(o_dir):
                os.makedirs(o_dir)

            o_fname = self.config['output_name'] % name_info
            o_path = os.path.join(o_dir, o_fname + '.sc')
            if compress:
                o_path += '.xz'

        ## if dir is missing make it
        dirname = os.path.dirname(o_path)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)

        if compress:
            assert o_path.endswith('.xz'), o_path
            logger.info('compress_and_encrypt_path(%r, tmp_dir=%r)', 
                        t_path, self.config['tmp_dir_path'])

            ## forcibly collect dereferenced objects
            #gc.collect()

            errors, t_path2 = streamcorpus.compress_and_encrypt_path(
                t_path, tmp_dir=self.config['tmp_dir_path'])
            assert not errors, errors

            if self.config.get('cleanup_tmp_files', True):
                # default action, move tmp file to output position
                try:
                    logger.debug('attempting renamed(%r, %r)', t_path2, o_path)
                    os.rename(t_path2, o_path)
                    logger.debug('renamed(%r, %r)', t_path2, o_path)
                except OSError, exc:
                    if exc.errno==18:
                        logger.debug('resorting to patient_move(%r, %r)',
                                     t_path2, o_path, exc_info=True)
                        patient_move(t_path2, o_path)
                        logger.debug('patient_move succeeded')
                    else:
                        logger.critical('rename failed (%r -> %r)', t_path2, o_path, exc_info=True)
                        raise
                return o_path
            else:
                # for debugging, leave temp file, copy to output
                shutil.copy(t_path2, o_path)
                logger.info('copied %r -> %r', t_path2, o_path)
                return o_path

        if self.config.get('cleanup_tmp_files', True):
            ## do an atomic renaming
            try:
                logger.debug('attemping os.rename(%r, %r)', t_path, o_path)
                os.rename(t_path, o_path)
            except OSError, exc:                
                if exc.errno==18:
                    patient_move(t_path, o_path)
                else:
                    logger.critical(
                        'failed shutil.copy2(%r, %r) and/or os.remove(t_path)',
                        t_path, o_path, exc_info=True)
                    raise
            except Exception, exc:
                logger.critical('failed os.rename(%r, %r)', t_path, o_path, exc_info=True)
                raise
        else:
            # for debugging, leave the tmp file, copy to output position
            shutil.copy(t_path, o_path)
            ogger.info('copied %r -> %r', t_path, o_path)

        ## return the final output path
        return o_path


class to_local_tarballs(object):
    def __init__(self, config):
        self.config = config

    def __call__(self, t_path, name_info, i_str):
        name_info.update( get_name_info( t_path, i_str=i_str ) )

        if name_info['num'] == 0:
            return None

        o_fname = self.config['output_name'] % name_info
        o_dir = self.config['output_path']
        o_path = os.path.join(o_dir, o_fname + '.tar.gz')

        ## if dir is missing make it
        dirname = os.path.dirname(o_path)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)

        t_path2 = tarball_export(t_path, name_info)

        ## do an atomic renaming    
        try:
            logger.debug('attemping os.rename(%r, %r)' % (t_path2, o_path))
            os.rename(t_path2, o_path)
        except OSError, exc:                
            if exc.errno==18:
                patient_move(t_path2, o_path)
            else:
                msg = 'failed shutil.copy2(%r, %r) and/or os.remove(t_path)\n%s'\
                    % (t_path2, o_path, traceback.format_exc(exc))
                logger.critical(traceback.format_exc(exc))
                raise
        except Exception, exc:
            msg = 'failed os.rename(%r, %r) -- %s' % (t_path, o_path, traceback.format_exc(exc))
            logger.critical(msg)
            raise

        try:
            os.remove(t_path)
        except Exception, exc:
            logger.critical('failed to os.remove(%r) --> %s' % (t_path, exc))

        logger.info('to_local_tarballs created %s' % o_path)

        ## return the final output path
        return o_path
